#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* somewhat unix-specific */
#include <sys/time.h>
#include <unistd.h>

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>


#include "my-curl.h"
/* #include "jhash.h" */


/*
 *  Checking the code with splint:
 *  clear && splint -unrecog -nullpass -nullret my-curl.c | less
 *
 */

/*____________________________________________________________________*/
inline size_t
WriteMemoryCallbackStruct(void *ptr, size_t size, size_t nmemb, void *data)
{
  /*register int realsize = size * nmemb;*/
  register size_t realsize = size * nmemb;

  result_t *mem = (result_t *)data;

  mem->memory = (unsigned char *)realloc(mem->memory, mem->size + realsize + 1);
  if (mem->memory) {
    memcpy(&(mem->memory[mem->size]), ptr, realsize);
    mem->size += realsize;
    mem->memory[mem->size] = '0';
  }
  return realsize;
}


/*____________________________________________________________________*/

inline CURL*
init(result_t *chunk_ptr)
{
  CURL *curl= curl_easy_init();

  assert(curl != NULL);
  
  /* DEBUG FUNCTION */
    /* NOT FAIL ON ERROR */
  curl_easy_setopt(curl, CURLOPT_FAILONERROR, 0);

  curl_easy_setopt(curl, CURLOPT_MUTE, 1);
     
  curl_easy_setopt(curl, CURLOPT_VERBOSE, 0);
  //curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, callback);

  curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);

  /* follow location */
  curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);

    /* send all data to this function  */
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallbackStruct);

  curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);

  
  curl_easy_setopt(curl, CURLOPT_DNS_CACHE_TIMEOUT, -1);


  curl_easy_setopt(curl, CURLOPT_DNS_USE_GLOBAL_CACHE, 1);

  curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, 99999999);

  curl_easy_setopt(curl, CURLOPT_TIMEOUT, 3); /* Server connection  time */

  curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 5); /* Transfert connection time */
    
  curl_easy_setopt(curl, CURLOPT_LOW_SPEED_LIMIT, 10);

  curl_easy_setopt(curl, CURLOPT_LOW_SPEED_TIME, 3);
  
  curl_easy_setopt(curl, CURLOPT_MAXFILESIZE, 102800); /*102 * 1024 = 102800bytes = 100Ko*/

  curl_easy_setopt(curl, CURLOPT_MAXCONNECTS, 30);

  curl_easy_setopt(curl, CURLOPT_ENCODING, "gzip");

  /* Curl error buffer with size = CURL_ERROR_SIZE */
  curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, chunk_ptr->error);

  /* we pass our 'chunk_ptr' struct to the callback function */
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *) chunk_ptr);

  return curl;
}


/*____________________________________________________________________*/

inline result_t * createMemoryStruct(void)
{
  result_t *chunk_ptr =NULL;

  chunk_ptr =(result_t *) malloc(sizeof (result_t));

  assert(chunk_ptr != NULL);

  chunk_ptr->memory =NULL;

  chunk_ptr->error  =(unsigned char *) malloc(sizeof (char) * CURL_ERROR_SIZE); 

  chunk_ptr->size            =0;    /* no data at this point */
  chunk_ptr->effective_url   =NULL;
  chunk_ptr->mime_type       =NULL;
  chunk_ptr->http_code       =0;
  chunk_ptr->curl_error_code =0;

  return chunk_ptr;
}


/*____________________________________________________________________*/
void
cleanup_memory(result_t *chunk_ptr)
{
  if(chunk_ptr->memory != NULL)
    {
      free(chunk_ptr->memory);
      chunk_ptr->memory =NULL;
    } 

  chunk_ptr->size            =0;
  chunk_ptr->http_code       =0;
  chunk_ptr->curl_error_code =0;
}

/*____________________________________________________________________*/
inline void
setError(result_t *chunk_ptr, int res, unsigned char *url)
{
  /* Set when errors occurs */
  /* Free the last allocated memory */
  if(chunk_ptr->memory != NULL)
    {
      free(chunk_ptr->memory);
      chunk_ptr->memory =NULL;
    }
  
  chunk_ptr->curl_error_code =res;
  chunk_ptr->http_code       =res;
  chunk_ptr->effective_url   =url;
  chunk_ptr->mime_type       =chunk_ptr->error;
}


/*____________________________________________________________________*/
inline long
getHTTPCode(result_t *chunk)
{
  return chunk->http_code;
}


/*____________________________________________________________________*/
inline unsigned char*
getEffectiveUrl(result_t *chunk)
{
  return chunk->effective_url; 
}


/*____________________________________________________________________*/
inline unsigned char*
getBody(result_t *chunk)
{
  if(chunk != NULL && chunk->curl_error_code == 0 && chunk->memory != NULL)
    {
      if(chunk->memory[chunk->size] != '\0')
        chunk->memory[chunk->size] = '\0';
      return chunk->memory;
    }
  else
    {
      return NULL;
    }
}


/*____________________________________________________________________*/
inline unsigned char*
getMimeType(result_t *chunk)
{
  return  chunk->mime_type;
}

/*____________________________________________________________________*/
inline int
getBodySize(result_t *chunk)
{
  return  chunk->size;
}
/*____________________________________________________________________*/

inline void
cleanup_chunk(result_t *chunk_ptr)
{
  if(chunk_ptr != NULL)
    { 
      cleanup_memory(chunk_ptr);

      if(chunk_ptr->error != NULL)
        {
          free(chunk_ptr->error);
          chunk_ptr->error =NULL;
        } 

      free(chunk_ptr);
      chunk_ptr =NULL;
    }
}


/*____________________________________________________________________*/

inline void
cleanup(CURL *curl)
{
  /* always cleanup */
  if(curl != NULL)
    {
      curl_easy_cleanup(curl);
      curl =NULL;
    }
}

/*____________________________________________________________________*/


int main(int argc, char **argv)
{
  char *sites[] =
    {
      "http://www.w3c.org",
      "http://search.cpan.org/recent"
    };

  int HANDLECOUNT =sizeof(sites)/sizeof(sites[0]); 

  result_t *chunks[HANDLECOUNT];
  CURL     *handles[HANDLECOUNT];

  CURLM    *multi_handle;

  int still_running; /* keep number of running handles */
  int i, fd;

  CURLMsg *msg; /* for picking up messages with the transfer status */
  int msgs_left; /* how many messages are left */

  struct timeval timeout;
  int rc; /* select() return code */
  int maxfd;

  result_t *input_chunk;
  CURL *input_handle;

  /* init a multi stack */
  multi_handle = curl_multi_init();

  /* Allocate one chunk per transfer */
  input_chunk =createMemoryStruct();

  /* Allocate one CURL handle per transfer */
  input_handle = init(input_chunk);

  for (i=0; i < HANDLECOUNT; i++)
    {
      if((handles[i] =curl_easy_duphandle(input_handle)) != NULL)
        {
          /* set the options (I left out a few, you'll get the point anyway) */
          curl_easy_setopt(handles[i], CURLOPT_URL, sites[i]);

          /* add the individual transfers */
          curl_multi_add_handle(multi_handle, handles[i]);
          //printf(">>>>> handles[%d] created\n", i);
        }
      else
        {
          printf(">>>>> curl_easy_duphandle error when [%d] created\n", i);
          exit(1);
        }
    }

  /* we start some action by calling perform right away */
  while(CURLM_CALL_MULTI_PERFORM == curl_multi_perform(multi_handle, &still_running));

  while(still_running)
    {
      fd_set fdread;
      fd_set fdwrite;
      fd_set fdexcep;

      FD_ZERO(&fdread);
      FD_ZERO(&fdwrite);
      FD_ZERO(&fdexcep);

      /* set a suitable timeout to play around with */
      timeout.tv_sec  = 1;
      timeout.tv_usec = 0;

      /* get file descriptors from the transfers */
      curl_multi_fdset(multi_handle, &fdread, &fdwrite, &fdexcep, &maxfd);

      rc = select(maxfd+1, &fdread, &fdwrite, &fdexcep, &timeout);

      switch(rc)
        {
        case -1:
          /* select error */
          
          /* Set when errors occurs */
          /* Free the last allocated memory */
          /*setError(chunks[i], res, url);*/
          perror(-1);
          break;
        case 0:
          /* timeout */

          /* Set when errors occurs */
          /* Free the last allocated memory */

          break;

        default:
          /* timeout or readable/writable sockets */
          while(CURLM_CALL_MULTI_PERFORM == curl_multi_perform(multi_handle, &still_running));
          
          break;
        }
    }


  /* See how the transfers went */
  while ((msg = curl_multi_info_read(multi_handle, &msgs_left)))
    {
      if (msg->msg == CURLMSG_DONE)
        {
          int idx, found = 0;
        
          /* Find out which handle this message is about */
          for (idx=0; (!found && (idx < HANDLECOUNT)); idx++)
            found = (msg->easy_handle == handles[idx]);

          for (i=0; i < HANDLECOUNT; i++)
            {
              if (idx == i)
                {
                  printf("[%d] Transfer completed with status %d\n", idx,msg->data.result);
                }
            }
        }
      else
        {
          printf("CURLMSG_NOT_DONE\n");
        }
    }

  /* cleanup memory */
  curl_multi_cleanup(multi_handle);

  /* Gets Results */
  for (i=0; i<HANDLECOUNT; i++)
    {
      /* Get the Curlcode */
      curl_easy_getinfo(handles[i], CURLINFO_HTTP_CODE, &(chunks[i]->http_code));
      /* Get the effective URL */
      curl_easy_getinfo(handles[i], CURLINFO_EFFECTIVE_URL, &(chunks[i]->effective_url));

      /* Get the mime type of the content */
      curl_easy_getinfo(handles[i], CURLINFO_CONTENT_TYPE, &(chunks[i]->mime_type));
    }

  /* Print Results */
  for (i=0; i<HANDLECOUNT; i++)
    {
      char *body = NULL;

      printf("URL %d :\n", i);
      printf("\tHTTP CODE     : %ld\n", getHTTPCode(chunks[i]));
      printf("\tEFFECTIVE URL : %s\n",  getEffectiveUrl(chunks[i]));
      printf("\tMIME TYPE     : %s\n",  getMimeType(chunks[i]));
      if((body = getBody(chunks[i])) != NULL)
        {
          printf("\tBODY SIZE     : %d\n", getBodySize(chunks[i]));
          /*printf("\tBODY          : %s\n", body);*/
        }
      else
        printf("\tNO BODY\n");
      
      printf( "\n");
    }

  /* Free the CURL handles */
  for (i=0; i<HANDLECOUNT; i++)
    {
      /* curl_easy_cleanup(handles[i]); */
      cleanup_chunk(chunks[i]);
      cleanup(handles[i]);
    }

  /* curl_easy_cleanup(input_chunk); */
  cleanup_chunk(input_chunk);
  cleanup(input_handle);


}



